/*
 *  Example 29: Shadow Volumes
 *
 *  This code uses one coordinate system for both the light and
 *  the objects to make the shadow volume calculations simple.
 *  This is obviously not real smart, but it makes things easy.
 *
 *  'm' Shadow volume variation
 *  'o' Objects
 *  +/- Light elevation
 *  []  Light position
 *  's' Start/stop light animation
 *  <>  Number of slices in objects
 *  'a' Toggle axes
 *  '0' snaps angles to 0,0
 *  arrows to rotate the world
 *  PgUp/PgDn zooms in/out
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
//  OpenGL with prototypes for glext
#define GL_GLEXT_PROTOTYPES
#include <GL/glut.h>
int mode=2;    //  Display mode
int obj=0;     //  Display objects
int move=1;    //  Light movement
int axes=1;    //  Display axes
int n =  8;    //  Number of slices
int th=-30;    //  Azimuth of view angle
int ph=+30;    //  Elevation of view angle
int inf=8;     //  Our sorry excuse for infinity
double asp=1;  //  Aspect ratio
double dim=3;  //  Size of world
typedef struct {float x,y,z;} Point;
typedef struct {float x,y,z,w;} Point4;
// Light values
int zh       =   0;  // Light azimuth
float Ylight =   2;  // Elevation of light
Point4 Lpos  = {1,1,1,1};  // Light position
//  Modes
char* text[] ={"Front Shadows","Back Shadows","Z-pass","Z-fail"};

#define MAXN 64
#define Dfloor  8
#define Yfloor -2

//  Macro for sin & cos in degrees
#define Cos(th) cos(3.1415927/180*(th))
#define Sin(th) sin(3.1415927/180*(th))

//  Rim, body, lid, and bottom data must be reflected in x and y;
//  handle and spout data across the y axis only.
static int patch[][4][4] =
{

   {{102,103,104,105},{  4,  5,  6,  7},{  8,  9, 10, 11},{ 12, 13, 14, 15}},  //  Rim
   {{ 12, 13, 14, 15},{ 16, 17, 18, 19},{ 20, 21, 22, 23},{ 24, 25, 26, 27}},  //  Upper body
   {{ 24, 25, 26, 27},{ 29, 30, 31, 32},{ 33, 34, 35, 36},{ 37, 38, 39, 40}},  //  Lower body
   {{118,118,118,118},{124,122,119,121},{123,126,125,120},{ 40, 39, 38, 37}},  //  Bottom
   {{ 96, 96, 96, 96},{ 97, 98, 99,100},{101,101,101,101},{  0,  1,  2,  3}},  //  Lid handle
   {{  0,  1,  2,  3},{106,107,108,109},{110,111,112,113},{114,115,116,117}},  //  Lid
   {{ 41, 42, 43, 44},{ 45, 46, 47, 48},{ 49, 50, 51, 52},{ 53, 54, 55, 56}},  //  Upper handle
   {{ 53, 54, 55, 56},{ 57, 58, 59, 60},{ 61, 62, 63, 64},{ 28, 65, 66, 67}},  //  Lower handle
   {{ 68, 69, 70, 71},{ 72, 73, 74, 75},{ 76, 77, 78, 79},{ 80, 81, 82, 83}},  //  Spout body
   {{ 80, 81, 82, 83},{ 84, 85, 86, 87},{ 88, 89, 90, 91},{ 92, 93, 94, 95}},  //  Spout tip
   {{127,128,129,130},{131,132,133,134},{135,136,137,138},{139,140,141,142}},  //  Torus outside
   {{143,144,145,146},{147,148,149,150},{151,152,153,154},{155,156,157,158}},  //  Torus inside
};

//  Data points
static Point data[] =
{
   { 0.2   ,  0     , 2.7    },
   { 0.2   , -0.112 , 2.7    },
   { 0.112 , -0.2   , 2.7    },
   { 0     , -0.2   , 2.7    },
   { 1.3375,  0     , 2.53125},
   { 1.3375, -0.749 , 2.53125},
   { 0.749 , -1.3375, 2.53125},
   { 0     , -1.3375, 2.53125},
   { 1.4375,  0     , 2.53125},
   { 1.4375, -0.805 , 2.53125},
   { 0.805 , -1.4375, 2.53125},
   { 0     , -1.4375, 2.53125},
   { 1.5   ,  0     , 2.4    },
   { 1.5   , -0.84  , 2.4    },
   { 0.84  , -1.5   , 2.4    },
   { 0     , -1.5   , 2.4    },
   { 1.75  ,  0     , 1.875  },
   { 1.75  , -0.98  , 1.875  },
   { 0.98  , -1.75  , 1.875  },
   { 0     , -1.75  , 1.875  },
   { 2     ,  0     , 1.35   },
   { 2     , -1.12  , 1.35   },
   { 1.12  , -2     , 1.35   },
   { 0     , -2     , 1.35   },
   { 2     ,  0     , 0.9    },
   { 2     , -1.12  , 0.9    },
   { 1.12  , -2     , 0.9    },
   { 0     , -2     , 0.9    },
   { -2    ,  0     , 0.9    },
   { 2     ,  0     , 0.45   },
   { 2     , -1.12  , 0.45   },
   { 1.12  , -2     , 0.45   },
   { 0     , -2     , 0.45   },
   { 1.5   ,  0     , 0.225  },
   { 1.5   , -0.84  , 0.225  },
   { 0.84  , -1.5   , 0.225  },
   { 0     , -1.5   , 0.225  },
   { 1.5   ,  0     , 0.15   },
   { 1.5   , -0.84  , 0.15   },
   { 0.84  , -1.5   , 0.15   },
   { 0     , -1.5   , 0.15   },
   {-1.6   ,  0     , 2.025  },
   {-1.6   , -0.3   , 2.025  },
   {-1.5   , -0.3   , 2.25   },
   {-1.5   ,  0     , 2.25   },
   {-2.3   ,  0     , 2.025  },
   {-2.3   , -0.3   , 2.025  },
   {-2.5   , -0.3   , 2.25   },
   {-2.5   ,  0     , 2.25   },
   {-2.7   ,  0     , 2.025  },
   {-2.7   , -0.3   , 2.025  },
   {-3     , -0.3   , 2.25   },
   {-3     ,  0     , 2.25   },
   {-2.7   ,  0     , 1.8    },
   {-2.7   , -0.3   , 1.8    },
   {-3     , -0.3   , 1.8    },
   {-3     ,  0     , 1.8    },
   {-2.7   ,  0     , 1.575  },
   {-2.7   , -0.3   , 1.575  },
   {-3     , -0.3   , 1.35   },
   {-3     ,  0     , 1.35   },
   {-2.5   ,  0     , 1.125  },
   {-2.5   , -0.3   , 1.125  },
   {-2.65  , -0.3   , 0.9375 },
   {-2.65  ,  0     , 0.9375 },
   {-2     , -0.3   , 0.9    },
   {-1.9   , -0.3   , 0.6    },
   {-1.9   ,  0     , 0.6    },
   { 1.7   ,  0     , 1.425  },
   { 1.7   , -0.66  , 1.425  },
   { 1.7   , -0.66  , 0.6    },
   { 1.7   ,  0     , 0.6    },
   { 2.6   ,  0     , 1.425  },
   { 2.6   , -0.66  , 1.425  },
   { 3.1   , -0.66  , 0.825  },
   { 3.1   ,  0     , 0.825  },
   { 2.3   ,  0     , 2.1    },
   { 2.3   , -0.25  , 2.1    },
   { 2.4   , -0.25  , 2.025  },
   { 2.4   ,  0     , 2.025  },
   { 2.7   ,  0     , 2.4    },
   { 2.7   , -0.25  , 2.4    },
   { 3.3   , -0.25  , 2.4    },
   { 3.3   ,  0     , 2.4    },
   { 2.8   ,  0     , 2.475  },
   { 2.8   , -0.25  , 2.475  },
   { 3.525 , -0.25  , 2.49375},
   { 3.525 ,  0     , 2.49375},
   { 2.9   ,  0     , 2.475  },
   { 2.9   , -0.15  , 2.475  },
   { 3.45  , -0.15  , 2.5125 },
   { 3.45  ,  0     , 2.5125 },
   { 2.8   ,  0     , 2.4    },
   { 2.8   , -0.15  , 2.4    },
   { 3.2   , -0.15  , 2.4    },
   { 3.2   ,  0     , 2.4    },
   { 0     ,  0     , 3.15   },
   { 0.8   ,  0     , 3.15   },
   { 0.8   , -0.45  , 3.15   },
   { 0.45  , -0.8   , 3.15   },
   { 0     , -0.8   , 3.15   },
   { 0     ,  0     , 2.85   },
   { 1.4   ,  0     , 2.4    },
   { 1.4   , -0.784 , 2.4    },
   { 0.784 , -1.4   , 2.4    },
   { 0     , -1.4   , 2.4    },
   //  Stretch the lid 15% for a tight fit
   { 0.4600, 0.0000 , 2.5500 },
   { 0.4600,-0.2576 , 2.5500 },
   { 0.2576,-0.4600 , 2.5500 },
   { 0.0000,-0.4600 , 2.5500 },
   { 1.4950, 0.0000 , 2.5500 },
   { 1.4950,-0.8372 , 2.5500 },
   { 0.8372,-1.4950 , 2.5500 },
   { 0.0000,-1.4950 , 2.5500 },
   { 1.4950, 0.0000 , 2.4000 },
   { 1.4950,-0.8372 , 2.4000 },
   { 0.8372,-1.4950 , 2.4000 },
   { 0.0000,-1.4950 , 2.4000 },
   { 0     ,  0     , 0      },
   { 1.425 , -0.798 , 0      },
   { 1.5   ,  0     , 0.075  },
   { 1.425 ,  0     , 0      },
   { 0.798 , -1.425 , 0      },
   { 0     , -1.5   , 0.075  },
   { 0     , -1.425 , 0      },
   { 1.5   , -0.84  , 0.075  },
   { 0.84  , -1.5   , 0.075  },
   // Add torus
   { 1.000 , 0.000  , 0.200  },
   { 1.300 , 0.000  , 0.200  },
   { 1.300 , 0.000  ,-0.200  },
   { 1.000 , 0.000  ,-0.200  },
   { 0.996 , 0.575  , 0.200  },
   { 1.295 , 0.747  , 0.200  },
   { 1.295 , 0.747  ,-0.200  },
   { 0.996 , 0.575  ,-0.200  },
   { 0.575 , 0.996  , 0.200  },
   { 0.748 , 1.295  , 0.200  },
   { 0.748 , 1.295  ,-0.200  },
   { 0.575 , 0.996  ,-0.200  },
   { 0.000 , 1.000  , 0.200  },
   { 0.000 , 1.300  , 0.200  },
   { 0.000 , 1.300  ,-0.200  },
   { 0.000 , 1.000  ,-0.200  },
   { 1.000 , 0.000  ,-0.200  },
   { 0.750 , 0.000  ,-0.200  },
   { 0.750 , 0.000  , 0.200  },
   { 1.000 , 0.000  , 0.200  },
   { 0.996 , 0.575  ,-0.200  },
   { 0.747 , 0.431  ,-0.200  },
   { 0.747 , 0.431  , 0.200  },
   { 0.996 , 0.575  , 0.200  },
   { 0.575 , 0.996  ,-0.200  },
   { 0.431 , 0.747  ,-0.200  },
   { 0.431 , 0.747  , 0.200  },
   { 0.575 , 0.996  , 0.200  },
   { 0.000 , 1.000  ,-0.200  },
   { 0.000 , 0.750  ,-0.200  },
   { 0.000 , 0.750  , 0.200  },
   { 0.000 , 1.000  , 0.200  },
   };

#define Bezier(x)  V*V*V*(U*U*U*p[0][0].x + 3*U*U*u*p[0][1].x + 3*U*u*u*p[0][2].x + u*u*u*p[0][3].x) \
               + 3*V*V*v*(U*U*U*p[1][0].x + 3*U*U*u*p[1][1].x + 3*U*u*u*p[1][2].x + u*u*u*p[1][3].x) \
               + 3*V*v*v*(U*U*U*p[2][0].x + 3*U*U*u*p[2][1].x + 3*U*u*u*p[2][2].x + u*u*u*p[2][3].x) \
               +   v*v*v*(U*U*U*p[3][0].x + 3*U*U*u*p[3][1].x + 3*U*u*u*p[3][2].x + u*u*u*p[3][3].x)
/*
 *  Evaluate 2D Bezier surface
 */
Point Bezier2D(Point p[4][4],float u,float v)
{
   float U = 1-u;
   float V = 1-v;
   Point P;
   P.x = Bezier(x);
   P.y = Bezier(y);
   P.z = Bezier(z);
   return P;
}

#define ddu(x)  -U*U*(V*V*V*p[0][0].x + 3*V*V*v*p[1][0].x + 3*V*v*v*p[2][0].x + v*v*v*p[3][0].x) \
         + (1-3*u)*U*(V*V*V*p[0][1].x + 3*V*V*v*p[1][1].x + 3*V*v*v*p[2][1].x + v*v*v*p[3][1].x) \
         + u*(2-3*u)*(V*V*V*p[0][2].x + 3*V*V*v*p[1][2].x + 3*V*v*v*p[2][2].x + v*v*v*p[3][2].x) \
         +       u*u*(V*V*V*p[0][3].x + 3*V*V*v*p[1][3].x + 3*V*v*v*p[2][3].x + v*v*v*p[3][3].x)
#define ddv(x)  -V*V*(U*U*U*p[0][0].x + 3*U*U*u*p[0][1].x + 3*U*u*u*p[0][2].x + u*u*u*p[0][3].x) \
         + (1-3*v)*V*(U*U*U*p[1][0].x + 3*U*U*u*p[1][1].x + 3*U*u*u*p[1][2].x + u*u*u*p[1][3].x) \
         + v*(2-3*v)*(U*U*U*p[2][0].x + 3*U*U*u*p[2][1].x + 3*U*u*u*p[2][2].x + u*u*u*p[2][3].x) \
         +       v*v*(U*U*U*p[3][0].x + 3*U*U*u*p[3][1].x + 3*U*u*u*p[3][2].x + u*u*u*p[3][3].x)

/*
 *  Evaluate 2D Bezier normal
 */
Point Normal2D(Point p[4][4],float u,float v)
{
   float tiny=1e-6;
   float U = 1-u;
   float V = 1-v;
   float D,Du,Dv;
   Point P,Pu,Pv;
   //  1/3 of derivative in the u direction
   Pu.x = ddu(x);
   Pu.y = ddu(y);
   Pu.z = ddu(z);
   Du = sqrt(Pu.x*Pu.x+Pu.y*Pu.y+Pu.z*Pu.z);
   //  1/3 of derivative in the v direction
   Pv.x = ddv(x);
   Pv.y = ddv(y);
   Pv.z = ddv(z);
   Dv = sqrt(Pv.x*Pv.x+Pv.y*Pv.y+Pv.z*Pv.z);
   //  Du=0
   if (Du<tiny && Dv>tiny)
   {
      u += 0.001;
      U -= 0.001;
      Pu.x = ddv(x);
      Pu.y = ddv(y);
      Pu.z = ddv(z);
   }
   //  Dv=0
   else if (Dv<tiny && Du>tiny)
   {
      v += 0.001;
      V -= 0.001;
      Pv.x = ddu(x);
      Pv.y = ddu(y);
      Pv.z = ddu(z);
   }
   //  Cross product
   P.x = Pu.y*Pv.z - Pu.z*Pv.y;
   P.y = Pu.z*Pv.x - Pu.x*Pv.z;
   P.z = Pu.x*Pv.y - Pu.y*Pv.x;
   //  Normalize
   D = sqrt(P.x*P.x+P.y*P.y+P.z*P.z);
   if (D>tiny)
   {
      P.x /= D;
      P.y /= D;
      P.z /= D;
   }
   return P;
}

/*
 *  Draw patch
 */
static void Patch(int I,float Sx,float Sy,float Sz,float x0,float y0,float z0,int light)
{
   int   i,j,k;
   int   V[MAXN+1][MAXN+1];
   Point p[4][4],P[MAXN+1][MAXN+1],N[MAXN+1][MAXN+1],T[MAXN+1][MAXN+1];

   //  Copy data with reflection
   for (k=0;k<4;k++)
   {
      int K = Sx*Sy*Sz<0 ? 3-k : k;
      for (j=0;j<4;j++)
      {
         int l = patch[I][j][K];
         p[j][k].x = x0+Sx*data[l].x;
         p[j][k].z = y0-Sy*data[l].y;
         p[j][k].y = z0+Sz*data[l].z;
      }
   }

   //  Evaluate vertexes
   for (i=0;i<=n;i++)
      for (j=0;j<=n;j++)
      {
         float u = (float)i/n;
         float v = (float)j/n;
         //  Vertex coordinates
         P[i][j] = Bezier2D(p,u,v);
         //  Normal
         N[i][j] = Normal2D(p,u,v);
         //  Texture coordinates
         T[i][j].x = 1-u;
         T[i][j].y = 1-v;
         T[i][j].z = 0;
         //  Is it visible from the light?
         V[i][j]= (N[i][j].x*(Lpos.x-P[i][j].x) + N[i][j].y*(Lpos.y-P[i][j].y) + N[i][j].z*(Lpos.z-P[i][j].z) >= 0);
      }

   //  Draw quads
   for (i=0;i<n;i++)
      for (j=0;j<n;j++)
         //  Draw quads making up object
         if (light)
         {
            glBegin(GL_QUADS);
            glNormal3fv((void*)&N[i  ][j  ]); glTexCoord2fv((void*)&T[i  ][j  ]); glVertex3fv((void*)&P[i  ][j  ]);
            glNormal3fv((void*)&N[i+1][j  ]); glTexCoord2fv((void*)&T[i+1][j  ]); glVertex3fv((void*)&P[i+1][j  ]);
            glNormal3fv((void*)&N[i+1][j+1]); glTexCoord2fv((void*)&T[i+1][j+1]); glVertex3fv((void*)&P[i+1][j+1]);
            glNormal3fv((void*)&N[i  ][j+1]); glTexCoord2fv((void*)&T[i  ][j+1]); glVertex3fv((void*)&P[i  ][j+1]);
            glEnd();
         }
         //  Draw shadow volume for those facets facing the light
         else if (V[i][j] | V[i+1][j] | V[i+1][j+1] | V[i][j+1])
         {
            //  Object coordinates (in a more convenient array)
            Point O[4] = {P[i][j],P[i+1][j],P[i+1][j+1],P[i][j+1]};
            //  Shadow coordinates (at infinity)
            Point S[4];
            for (k=0;k<4;k++)
            {
               S[k].x = inf*(O[k].x-Lpos.x) + O[k].x;
               S[k].y = inf*(O[k].y-Lpos.y) + O[k].y;
               S[k].z = inf*(O[k].z-Lpos.z) + O[k].z;
            }
            //  Front face
            glBegin(GL_QUADS);
            glVertex3f(O[0].x,O[0].y,O[0].z);
            glVertex3f(O[1].x,O[1].y,O[1].z);
            glVertex3f(O[2].x,O[2].y,O[2].z);
            glVertex3f(O[3].x,O[3].y,O[3].z);
            //  Back face
            glVertex3f(S[3].x,S[3].y,S[3].z);
            glVertex3f(S[2].x,S[2].y,S[2].z);
            glVertex3f(S[1].x,S[1].y,S[1].z);
            glVertex3f(S[0].x,S[0].y,S[0].z);
            glEnd();
            //  Sides
            glBegin(GL_QUAD_STRIP);
            for (k=0;k<=4;k++)
            {
               glVertex3f(O[k%4].x,O[k%4].y,O[k%4].z);
               glVertex3f(S[k%4].x,S[k%4].y,S[k%4].z);
            }
            glEnd();
         }
}

/*
 *  Draw torus
 */
static void Torus(float x0,float y0,float z0,float S,int light)
{
   int i;
   for (i=10;i<=11;i++)
   {
      Patch(i,+S,+S,S,x0,y0,z0,light);
      Patch(i,+S,-S,S,x0,y0,z0,light);
      Patch(i,-S,-S,S,x0,y0,z0,light);
      Patch(i,-S,+S,S,x0,y0,z0,light);
   }
}

/*
 *  Draw teapot
 */
static void Teapot(float x0,float y0,float z0,float S,int light)
{
   int i;

   //  Draw objects
   for (i=0;i<10;i++)
   {
      if (i<6)
      {
         Patch(i,+S,+S,S,x0,y0,z0,light);
         Patch(i,+S,-S,S,x0,y0,z0,light);
         Patch(i,-S,+S,S,x0,y0,z0,light);
         Patch(i,-S,-S,S,x0,y0,z0,light);
      }
      //  Draw patch reflected to 2 hemispheres
      else
      {
         Patch(i,+S,+S,S,x0,y0,z0,light);
         Patch(i,+S,-S,S,x0,y0,z0,light);
      }
   }
}

/*
 *  Convenience routine to output raster text
 *  Use VARARGS to make this more flexible
 */
#define LEN 8192  //  Maximum length of text string
void Print(const char* format , ...)
{
   char    buf[LEN];
   char*   ch=buf;
   va_list args;
   //  Turn the parameters into a character string
   va_start(args,format);
   vsnprintf(buf,LEN,format,args);
   va_end(args);
   //  Display the characters one at a time at the current raster position
   while (*ch)
      glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18,*ch++);
}

/*
 *  Set projection
 */
static void Project()
{
   //  Tell OpenGL we want to manipulate the projection matrix
   glMatrixMode(GL_PROJECTION);
   //  Undo previous transformations
   glLoadIdentity();
   //  Perspective transformation
   gluPerspective(60,asp,dim/10,10*dim);
   //  Switch to manipulating the model matrix
   glMatrixMode(GL_MODELVIEW);
   //  Undo previous transformations
   glLoadIdentity();
}
/*
 *  Enable lighting
 */
static void Light(int on)
{
   if (on)
   {
      float Ambient[]   = {0.3,0.3,0.3,1.0};
      float Diffuse[]   = {1,1,1,1};
      //  Enable lighting with normalization
      glEnable(GL_LIGHTING);
      glEnable(GL_NORMALIZE);
      //  glColor sets ambient and diffuse color materials
      glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
      glEnable(GL_COLOR_MATERIAL);
      //  Enable light 0
      glEnable(GL_LIGHT0);
      glLightfv(GL_LIGHT0,GL_AMBIENT ,Ambient);
      glLightfv(GL_LIGHT0,GL_DIFFUSE ,Diffuse);
      glLightfv(GL_LIGHT0,GL_POSITION,(void*)&Lpos);
   }
   else
      glDisable(GL_LIGHTING);
}

/*
 *  Draw scene
 *    light>0  lit colors
 *    light<0  shaded colors
 *    light=0  shadow volumes
 */
void Scene(int light)
{
   int i,j;
   if (light<0)
      glColor3f(0.5,0.5,0);
   else
      glColor3f(1,1,0);
   //  Draw teapot
   Teapot(-0.5,0,-0.5,0.25,light);
   //  Draw tori
   if (obj)
   {
      Torus(+0.5,0,-0.8,0.5,light);
      Torus(+0.5,0,+0.8,0.5,light);
   }

   //  The floor doesn't cast a shadow
   if (!light) return;

   //  Draw floor
   if (light<0)
      glColor3f(0.5,0.5,0.5);
   else
      glColor3f(1,1,1);
   glNormal3f(0,1,0);
   glEnable(GL_TEXTURE_2D);
   for (j=-Dfloor;j<Dfloor;j++)
   {
      glBegin(GL_QUAD_STRIP);
      for (i=-Dfloor;i<=Dfloor;i++)
      {
         glTexCoord2f(i,j); glVertex3f(i,Yfloor,j);
         glTexCoord2f(i,j+1); glVertex3f(i,Yfloor,j+1);
      }
      glEnd();
   }
   glDisable(GL_TEXTURE_2D);
}

/*
 *  OpenGL (GLUT) calls this routine to display the scene
 */
void display()
{
   //  Length of axes
   const double len=2.0;
   //  Eye position
   float Ex = -2*dim*Sin(th)*Cos(ph);
   float Ey = +2*dim        *Sin(ph);
   float Ez = +2*dim*Cos(th)*Cos(ph);
   //  Light position
   Lpos.x = Cos(zh);
   Lpos.y = Ylight;
   Lpos.z = Sin(zh);

   //  Erase the window and the depth buffer
   glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT|GL_STENCIL_BUFFER_BIT);
   //  Enable Z-buffering in OpenGL
   glEnable(GL_DEPTH_TEST);

   //  Set perspective
   glLoadIdentity();
   gluLookAt(Ex,Ey,Ez , 0,0,0 , 0,Cos(ph),0);

   //  Draw light position as sphere (still no lighting here)
   glColor3f(1,1,1);
   glPushMatrix();
   glTranslated(Lpos.x,Lpos.y,Lpos.z);
   glutSolidSphere(0.03,10,10);
   glPopMatrix();

   /*
    *  Draw scene using shadow volumes
    */
   //  Lighting is still off
   //  The color should be what the object looks like when in the shadow
   //  This sets the object depths in the Z-buffer
   Scene(-1);

   //  Make color buffer and Z buffer read-only
   glColorMask(0,0,0,0);
   glDepthMask(0);

   //  Enable stencil
   //  Always draw regardless of the stencil value
   glEnable(GL_STENCIL_TEST);
   glStencilFunc(GL_ALWAYS,0,0xFFFFFFFF);

   //  Enable face culling
   glEnable(GL_CULL_FACE);

   switch (mode)
   {
      //  Front shadows
      //  This is for demonstation purposes only
      case 0:
         glFrontFace(GL_CCW);
         glStencilOp(GL_KEEP,GL_KEEP,GL_INCR);
         Scene(0);
         break;
      //  Back shadows
      //  This is for demonstation purposes only
      case 1:
         glFrontFace(GL_CW);
         glStencilOp(GL_KEEP,GL_KEEP,GL_INCR);
         Scene(0);
         break;
      //  Z-pass variation
      //  Count from the eye to the object
      case 2:
         //  Draw only the front faces of the shadow volume
         //  Increment the stencil value on Z pass
         //  Depth and color buffers unchanged
         glFrontFace(GL_CCW);
         glStencilOp(GL_KEEP,GL_KEEP,GL_INCR);
         Scene(0);
         //  Draw only the back faces of the shadow volume
         //  Decrement the stencil value on Z pass
         //  Depth and color buffers unchanged
         glFrontFace(GL_CW);
         glStencilOp(GL_KEEP,GL_KEEP,GL_DECR);
         Scene(0);
         break;
      //  Z-fail variation
      //  Count from the object to infinity
      case 3:
         //  Draw only the back faces of the shadow volume
         //  Increment the stencil value on Z fail
         //  Depth and color buffers unchanged
         glFrontFace(GL_CW);
         glStencilOp(GL_KEEP,GL_INCR,GL_KEEP);
         Scene(0);
         //  Draw only the front faces of the shadow volume
         //  Decrement the stencil value on Z fail
         //  Depth and color buffers unchanged
         glFrontFace(GL_CCW);
         glStencilOp(GL_KEEP,GL_DECR,GL_KEEP);
         Scene(0);
         break;
      default:
         break;
   }

   //  Disable face culling
   glDisable(GL_CULL_FACE);

   //  Make color mask and depth buffer read-write
   glColorMask(1,1,1,1);

   //  Update the color only where the stencil value is 0
   //  Do not change the stencil
   glStencilFunc(GL_EQUAL,0,0xFFFFFFFF);
   glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
   //  Redraw only if depth matches front object
   glDepthFunc(GL_EQUAL);

   //  Enable lighting
   //  Now render the scene for real
   Light(1);
   Scene(1);

   /*
    *  Here endeth the shadow volumes
    */

   //  Undo changes
   //  Disable lighting
   Light(0);
   glDisable(GL_STENCIL_TEST);
   glDepthFunc(GL_LESS); 
   glDepthMask(1);

   //  Draw axes (white)
   glColor3f(1,1,1);
   if (axes)
   {
      glBegin(GL_LINES);
      glVertex3d(0.0,0.0,0.0);
      glVertex3d(len,0.0,0.0);
      glVertex3d(0.0,0.0,0.0);
      glVertex3d(0.0,len,0.0);
      glVertex3d(0.0,0.0,0.0);
      glVertex3d(0.0,0.0,len);
      glEnd();
      //  Label axes
      glRasterPos3d(len,0.0,0.0);
      Print("X");
      glRasterPos3d(0.0,len,0.0);
      Print("Y");
      glRasterPos3d(0.0,0.0,len);
      Print("Z");
   }

   //  Display parameters
   glWindowPos2i(5,5);
   Print("Ylight=%.1f Angle=%d,%d  Dim=%.1f Slices=%d Mode=%s Infinity=%d",
     Ylight,th,ph,dim,n,text[mode],inf);

   //  Render the scene and make it visible
   glFlush();
   glutSwapBuffers();
}

/*
 *  GLUT calls this routine when the window is resized
 */
void idle()
{
   //  Elapsed time in seconds
   double t = glutGet(GLUT_ELAPSED_TIME)/1000.0;
   zh = fmod(90*t,360.0);
   //  Tell GLUT it is necessary to redisplay the scene
   glutPostRedisplay();
}

/*
 *  GLUT calls this routine when an arrow key is pressed
 */
void special(int key,int x,int y)
{
   //  Right arrow key - increase angle by 5 degrees
   if (key == GLUT_KEY_RIGHT)
      th += 1;
   //  Left arrow key - decrease angle by 5 degrees
   else if (key == GLUT_KEY_LEFT)
      th -= 1;
   //  Up arrow key - increase elevation by 5 degrees
   else if (key == GLUT_KEY_UP)
      ph += 1;
   //  Down arrow key - decrease elevation by 5 degrees
   else if (key == GLUT_KEY_DOWN)
      ph -= 1;
   //  PageUp key - increase dim
   else if (key == GLUT_KEY_PAGE_DOWN)
      dim += 0.1;
   //  PageDown key - decrease dim
   else if (key == GLUT_KEY_PAGE_UP && dim>1)
      dim -= 0.1;
   //  Keep angles to +/-360 degrees
   th %= 360;
   ph %= 360;
   //  Update projection
   Project();
   //  Tell GLUT it is necessary to redisplay the scene
   glutPostRedisplay();
}

/*
 *  GLUT calls this routine when a key is pressed
 */
void key(unsigned char ch,int x,int y)
{
   //  Exit on ESC
   if (ch == 27)
      exit(0);
   //  Reset view angle
   else if (ch == '0')
      th = ph = 0;
   //  Toggle axes
   else if (ch == 'a' || ch == 'A')
      axes = 1-axes;
   //  Toggle display modes
   else if (ch == 'm')
      mode = (mode+1)%4;
   else if (ch == 'M')
      mode = (mode+3)%4;
   //  Toggle light movement
   else if (ch == 's' || ch == 'S')
      move = 1-move;
   //  Toggle objects
   else if (ch == 'o' || ch == 'O')
      obj = 1-obj;
   //  Light elevation
   else if (ch=='-')
      Ylight -= 0.1;
   else if (ch=='+')
      Ylight += 0.1;
   //  Light azimuth
   else if (ch=='[')
      zh -= 1;
   else if (ch==']')
      zh += 1;
   //  Number of patches
   else if (ch=='<' && n>1)
      n--;
   else if (ch=='>' && n<MAXN)
      n++;
   //  Infinity
   else if (ch=='I' && inf>2)
      inf /= 2;
   else if (ch=='i')
      inf *= 2;
   //  Set idle function
   glutIdleFunc(move?idle:NULL);
   //  Reproject
   Project();
   //  Tell GLUT it is necessary to redisplay the scene
   glutPostRedisplay();
}

/*
 *  GLUT calls this routine when the window is resized
 */
void reshape(int width,int height)
{
   //  Ratio of the width to the height of the window
   asp = (height>0) ? (double)width/height : 1;
   //  Set the viewport to the entire window
   glViewport(0,0, width,height);
   //  Set projection
   Project();
}

/*
 *  Print message to stderr and exit
 */
void fatal(const char* format , ...)
{
   va_list args;
   va_start(args,format);
   vfprintf(stderr,format,args);
   va_end(args);
   exit(1);
}
/*
 *  Reverse n bytes
 */
#ifdef BIGEND
static void reverse(void* x,const int n)
{
   int k;
   char* ch=x;
   for (k=0;k<n/2;k++)
   {
      char tmp = ch[k];
      ch[k] = ch[n-1-k];
      ch[n-1-k] = tmp;
   }
}
#endif
/*
 *  Load texture from BMP file
 */
unsigned int LoadTexBMP(char* file)
{
   unsigned int   texture;     // Texture name
   FILE*          f;           // File pointer
   char           magic[2];    // Image magic
   unsigned int   dx,dy,size;  // Image dimensions
   unsigned short nbp,bpp;     // Planes and bits per pixel
   unsigned char* image;       // Image data
   int            k;           // Counter

   //  Open file
   f = fopen(file,"rb");
   if (!f) fatal("Cannot open file %s\n",file);
   //  Check image magic
   if (fread(magic,2,1,f)!=1) fatal("Cannot read magic from %s\n",file);
   if (strncmp(magic,"BM",2)) fatal("Image magic not BMP in %s\n",file);
   //  Seek to and read header
   fseek(f,16,SEEK_CUR);
   if (fread(&dx ,4,1,f)!=1 || fread(&dy ,4,1,f)!=1 || fread(&nbp,2,1,f)!=1 || fread(&bpp,2,1,f)!=1)
     fatal("Cannot read header from %s\n",file);
#ifdef BIGEND
   //  Reverse bytes on big-endian hardware
   reverse(&dx,4);
   reverse(&dy,4);
   reverse(&nbp,2);
   reverse(&bpp,2);
#endif
   //  Check image parameters
   if (nbp!=1) fatal("%s bit planes is not 1: %d\n",file,nbp);
   if (bpp!=24) fatal("%s bits per pixel is not 24: %d\n",file,bpp);
#ifndef GL_VERSION_2_0
   //  OpenGL 2.0 lifts the restriction that texture size must be a power of two
   for (k=1;k<dx;k*=2);
   if (k!=dx) fatal("%s image width not a power of two: %d\n",file,dx);
   for (k=1;k<dy;k*=2);
   if (k!=dy) fatal("%s image height not a power of two: %d\n",file,dy);
#endif

   //  Allocate image memory
   size = 3*dx*dy;
   image = (unsigned char*) malloc(size);
   if (!image) fatal("Cannot allocate %d bytes of memory for image %s\n",size,file);
   //  Seek to and read image
   fseek(f,24,SEEK_CUR);
   if (fread(image,size,1,f)!=1) fatal("Error reading data from image %s\n",file);
   fclose(f);
   //  Reverse colors (BGR -> RGB)
   for (k=0;k<size;k+=3)
   {
      unsigned char temp = image[k];
      image[k]   = image[k+2];
      image[k+2] = temp;
   }

   //  Generate 2D texture
   glGenTextures(1,&texture);
   glBindTexture(GL_TEXTURE_2D,texture);
   //  Copy image
   glTexImage2D(GL_TEXTURE_2D,0,3,dx,dy,0,GL_RGB,GL_UNSIGNED_BYTE,image);
   if (glGetError()) fatal("Error in glTexImage2D %s %dx%d\n",file,dx,dy);
   //  Scale linearly when image size doesn't match
   glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);

   //  Free image memory
   free(image);
   //  Return texture name
   return texture;
}

/*
 *  Start up GLUT and tell it what to do
 */
int main(int argc,char* argv[])
{
   //  Initialize GLUT
   glutInit(&argc,argv);
   //  Request double buffered, true color window with Z buffering at 600x600
   glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE | GLUT_STENCIL);
   glutInitWindowSize(600,600);
   glutCreateWindow("Utah Teapot");
   //  Set callbacks
   glutDisplayFunc(display);
   glutReshapeFunc(reshape);
   glutSpecialFunc(special);
   glutKeyboardFunc(key);
   glutIdleFunc(move?idle:NULL);
   //  Floor Texture
   LoadTexBMP("water.bmp");
   //  Pass control to GLUT so it can interact with the user
   glutMainLoop();
   return 0;
}
